/*************************************************************************/
/*                                                                       */
/* Program: rdsss.c                                                      */
/*                                                                       */
/* Random Dot Stereogram program                                         */
/* =============================                                         */
/*                                                                       */
/* This program calculates a Random Dot Stereogram image.                */
/* This is an image with black dots on a white background                */
/* which looks noisy. But if you focus your eyes behind                  */
/* the image plane a 3-dimensional image will appear.                    */
/*                                                                       */
/* The program use Interval Bisection to perform ray-tracing.            */
/* (Newton Raphson is commented out for numerical reasons.)              */
/*                                                                       */
/* Algorithm and program by: Stefan Spaennare, Lund, Sweden              */
/* E-mail:                   stefan@spaennare.se                         */
/*                                                                       */
/* First version:            October 1992                                */
/* Latest update:            2006-01-01                                  */
/*                                                                       */
/*************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* Notice:                                                                   */
/* =======                                                                   */
/*                                                                           */
/* I make no warranties that this program is (1) free of error, (2) con-     */
/* sistent with any standard merchantability, or (3) meeting the require-    */
/* ments of a particular application.  This software shall not, partly or    */
/* as a whole, participate in a process, whose outcome can result in injury  */
/* to a person or loss of property. It is solely designed for analytical     */
/* work.  Permission to use, copy, and distribute is hereby granted without  */
/* fee, providing that the header above including this notice appears in     */
/* all copies.                                                               */
/*                                                                           */
/*                                                         Stefan Spaennare  */
/*                                                                           */
/*****************************************************************************/


#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include "image.h"


double timetic=(double)(CLOCKS_PER_SEC);

double pi=3.14159265358979323;

double p0,p1,p2,p3,p4;


#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

double ran2(idum)
long *idum;
{
   int j;
   long k;
   static long idum2=123456789;
   static long iy=0; 
   static long iv[NTAB];
   double temp;

   if (*idum <= 0) {
      if (-(*idum) < 1) {
         *idum=1;
      }
      else {
         *idum=-(*idum);
      } /* if */
      idum2=(*idum);
      for (j=NTAB+7; j>=0; j--) {
         k=(*idum)/IQ1;
	 *idum=IA1*(*idum-k*IQ1)-k*IR1;
	 if (*idum < 0) {
	    *idum += IM1;
	 } /* if */
	 if (j < NTAB) {
	    iv[j]=*idum;
	 } /* if */
      } /* for j */
      iy=iv[0];
   } /* if */
   k=(*idum)/IQ1;
   *idum=IA1*(*idum-k*IQ1)-k*IR1;
   if (*idum < 0) {
      *idum += IM1;
   } /* if */
   k=idum2/IQ2;
   idum2=IA2*(idum2-k*IQ2)-k*IR2;
   if (idum2 < 0) {
      idum2 += IM2;
   } /* if */
   j=iy/NDIV;
   iy=iv[j]-idum2;
   iv[j] = *idum;
   if (iy < 1) {
      iy += IMM1;
   } /* if */
   temp=AM*(double)(iy);
   if (temp > RNMX) {
      return(RNMX);
   }
   else {
      return(temp);
   } /* if */

} /* ran2 */


double f(x,y)
double x,y;
{
   double r,z,z1,x2,y2;

   z1=-p1;

   z=p0;

   x2=p3*x;
   y2=-p4*y;

   if (
      ((-35<x2) && (-5>x2) && (30>y2) && (20<y2)) ||
      ((-35<x2) && (-5>x2) && (5>y2) && (-5<y2)) ||
      ((-35<x2) && (-5>x2) && (-20>y2) && (-30<y2)) ||
      ((-35<x2) && (-25>x2) && (20>y2) && (5<y2)) ||
      ((-15<x2) && (-5>x2) && (-5>y2) && (-20<y2))||
      ((5<x2) && (35>x2) && (30>y2) && (20<y2)) ||
      ((5<x2) && (35>x2) && (5>y2) && (-5<y2)) ||
      ((5<x2) && (35>x2) && (-20>y2) && (-30<y2)) ||
      ((5<x2) && (15>x2) && (20>y2) && (5<y2)) ||
      ((25<x2) && (35>x2) && (-5>y2) && (-20<y2))) { 
      z=z1;
   } /* if */

   return(z);

} /* f */


double fl(x,y,d,d2,de,xl)
double x,y,d,d2,de,xl;
{
   return(d/(xl+de)*(x-xl)-d2-f(x,y));
} /* fl */


double fr(x,y,d,d2,de,xr)
double x,y,d,d2,de,xr;
{
   return(d/(xr-de)*(x-xr)-d2-f(x,y));
} /* fr */


int main(argc,argv)
int argc;
char *argv[];
{

struct image_head header;    
short **image;

double x0,y0,x,y,z,x2,xr,yr,xl,yl,d,d2,de,ax;
double epsib,eps,eps0,s,u,u2,q,a,b,n,nf,alfa,h; 
double flprim,frprim;

int i,j,k,xsr,ysr,xsl,ysl,xs,ys,kmax,rows,size,seed;

long idum;

clock_t time1,timediff1;


if (argc != 4) {
   printf("Usage: %s image sizei 3D-depth-of-objectr(10.0)\n",argv[0]);
   exit(0);
} /* if */


size=atoi(argv[2]);
p1=atof(argv[3]);


/* The parameters p0 to p4 are used as constants  */
/* in the function f that build up the 3D image.  */

p0=0.0;
p2=0.0;
p3=0.5;
p4=0.5;


d=200.0;  /* Distance from eyes to the image in mm. */
s=181.0;  /* Image size in mm. */

de=35.0;  /* 3D depth of the image in mm. */
ax=12.0;  /* Half distance between horisontal dots in mm. */

d2=d*ax/(de-ax);
u=(double)(size);
u2=u/2.0;
q=(u-1.0)/s;
n=52428.0/(512.0*512.0)*u*u;
nf=s/(2.0*ax);

rows=(int)(n/nf); /* Number of horisontal rows in the image. */


/*
epsib=0.01;
*/
epsib=1.0e-14;
eps=1.0e-15;
eps0=sqrt(eps);

kmax=10;


header.id=image_magic;
header.typ=im_s16;
header.off=512;
header.hgh=size;
header.wid=size;

image=(short **)calloc(header.hgh,sizeof(short*));
   
CreateIm_s16(image,header.hgh,header.wid);


for (i=0; i<size; i++) {
   for (j=0; j<size; j++) {
       image[j][i]=32767; 
   } /* for j */
} /* for i */


seed=3;
idum=-abs(seed)-1;

time1=clock();


for (i=0; i<rows; i++) {

   x0=s*(ran2(&idum)-0.5);
   y0=s*(ran2(&idum)-0.5);
   xs=(int)(q*x0+u2);
   ys=(int)(q*y0+u2);
   image[ys][xs]=0;

   y=(d+d2)/d*y0;

   xr=x0;
   x=x0;

   while (x < (0.6*s+2.0*ax)) {

      xl=xr;
      a=-s;
      b=s;
      x=1.0;

      while ((fabs(a-b)) > (fabs(epsib*x))) { 
         x=(a+b)/2.0;
         if ((fl(a,y,d,d2,de,xl)*
              fl(x,y,d,d2,de,xl)) > 0.0) {
            a=x;
         }
         else {
            b=x;
         } /* if */
      } /* while */

/*
      h=fabs(x)*eps0+eps0;
      k=0;

      do {
          x2=x;
          flprim=(fl(x+h,y,d,d2,de,xl)-
                  fl(x-h,y,d,d2,de,xl))/(2.0*h);
          x=x-fl(x,y,d,d2,de,xl)/flprim; 
          k++;
      } while (((fabs(x-x2)) > (fabs(x*eps))) && (k < kmax));
*/

      z=f(x,y);
      alfa=-(z+d2)/(z+d+d2);
      xr=x+(x-de)*alfa;
      xsr=(int)(q*xr+u2+0.5);
      ysr=(int)(q*y0+u2+0.5);

      if ((xsr >=0) && (xsr < size)) {
         if ((ysr >=0) && (ysr < size)) {
            image[ysr][xsr]=0;
         } /* if */
      } /* if */

   } /* while */

   xl=x0;
   x=x0;     

   while (x > (-0.6*s-2.0*ax)) {

      xr=xl;
      a=-s;
      b=s;
      x=1.0;

      while ((fabs(a-b)) > (fabs(epsib*x))) { 
         x=(a+b)/2.0;
         if ((fr(a,y,d,d2,de,xr)*
              fr(x,y,d,d2,de,xr)) > 0.0) {
            a=x;
         }
         else {
            b=x;
         } /* if */
      } /* while */

/*
      h=fabs(x)*eps0+eps0;
      k=0;

      do {
          x2=x;
          frprim=(fr(x+h,y,d,d2,de,xl)-
                  fr(x-h,y,d,d2,de,xl))/(2.0*h);
          x=x-fr(x,y,d,d2,de,xr)/frprim; 
          k++;
      } while (((fabs(x-x2)) > (fabs(x*eps))) && (k < kmax)); 
*/

      z=f(x,y);
      alfa=-(z+d2)/(z+d+d2);
      xl=x+(x+de)*alfa;
      xsl=(int)(q*xl+u2+0.5);
      ysl=(int)(q*y0+u2+0.5);

      if ((xsl >=0) && (xsl < size)) {
         if ((ysl >=0) && (ysl < size)) {
            image[ysl][xsl]=0;
         } /* if */
      } /* if */

   } /* while */

} /* for i */

timediff1=clock()-time1;

printf("\n");
printf("CPU-time: %8.2f s\n",(double)(timediff1)/timetic);
printf("\n");

WriteImage(argv[1],image,header,im_s16);

} /* End */
